// SilentTimer Action Figure adjust calibration value
// 
// 03.01.2021
// Copyright (C) 2021 Balthasar Szczepanski
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <stdio.h>
#include <stdint.h>
#include <inttypes.h>
#include "libstaf.h"

#define CLOCK_BASE (1966080) //2^15 * 60
#define CLOCK_MULT (-2)

double decodeCalibration(int16_t value);
int16_t encodeCalibration(double value);

double decodeCalibration(int16_t x)
{
	return (((double)x)*CLOCK_MULT+CLOCK_BASE)/CLOCK_BASE;
}

int16_t encodeCalibration(double x)
{
	double y = ((x*CLOCK_BASE)-CLOCK_BASE)/CLOCK_MULT;
	if(y >= 255.5)
		return 32767;
	if(y <= -255.5)
		return -32768;
	if(y >= 0)
		return (int16_t)(y+0.5);
	else
		return (int16_t)(y-0.5);
}

int main(int argc, char *argv[])
{
	int fd, r;
	
	double measured,actual;
	int16_t setting;
	
	struct StafState state;
	
	if (argc<2)
	{
		fprintf(stderr,"%s port [value]\n",(argc>0)?argv[0]:"staf-adjust");
		return 1;
	}
	
	fd = stafConnect(argv[1]);
	if (fd<1)
	{
		fprintf(stderr,"%d: Can't connect to %s.\n",fd,argv[1]);
		return fd;
	}
	
	r=stafGetState(fd,NULL,&state);
	if (r)
	{
		fprintf(stderr,"%d: Can't get STAF state.\n",r);
		stafDisconnect(fd);
		return r;
	}
	
	//1. get the measured time rate from command line or stdin
	if (argc>2)
		sscanf(argv[2],"%lf",&measured);
	else
		scanf("%lf",&measured);
	printf("measured clock rate: %lf\n",measured);
	
	//2. get current calibration setting
	r=stafGetCalibration(fd,NULL,&setting,0);
	if (r)
	{
		fprintf(stderr,"%d: Can't get STAF calibration setting.\n",r);
		stafDisconnect(fd);
		return r;
	}
	printf("current calibration setting: %"PRIi16"\n",setting);
	
	//3. determine actual clock rate
	actual=measured/decodeCalibration(setting);
	printf("actual clock rate: %lf\n",actual);
	
	//4. determine correct calibration setting
	setting=encodeCalibration(1.0/actual);
	if(setting>255)
	{
		setting=255;
		fprintf(stderr,"%d: Calibration value out of range; using maximum value.\n",r);
	}
	else if (setting<-255)
	{
		setting=-255;
		fprintf(stderr,"%d: Calibration value out of range, using minimum value.\n",r);
	}
	printf("new calibration setting: %"PRIi16"\n",setting);

	//5. apply new setting
	r=stafSetCalibration(fd,NULL,setting,0);
	if (r)
	{
		fprintf(stderr,"%d: Can't set STAF calibration.\n",r);
		stafDisconnect(fd);
		return r;
	}
	
	r=stafSetState(fd,NULL,&state);
	if (r)
	{
		fprintf(stderr,"%d: Can't restore STAF state.\n",r);
		stafDisconnect(fd);
		return r;
	}
	
	stafDisconnect(fd);
	return 0;
}
