// SilentTimer Action Figure resynchronise time with computer
// 
// 01.11.2022
// Copyright (C) 2022 Balthasar Szczepanski
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

// Will resynchronise to local time in your timezone, not UTC.

#include <time.h>
#include <stdio.h>
#include <stdint.h>
#include <inttypes.h>
#include <math.h>
#include "libstaf.h"

#define FILTER_FACTOR 0.02
#define STOP_CONDITION 0.001

#define STATE_INIT               0
#define STATE_WAIT_FULL_MINUTE   1
#define STATE_SET_TIME           2
#define STATE_WAIT_NEXT_MINUTE   3
#define STATE_WAIT_FILTER        4
#define STATE_SET_RECALIBRATION  5
#define STATE_RESYNCHRONISATION  6
#define STATE_STOP_RECALIBRATION 7
#define STATE_END                8

#define STATE_EXIT      255


int16_t calculate_factor(double diff);

int main(int argc, char *argv[])
{
	int fd, r, r2;
	int16_t calibration;
	uint8_t debug_val;
	
	struct StafState state;
	
	struct timespec comptime;
	struct tm staftm;
	struct tm *comptm;
	time_t staftime;
	
	double ts_staf, ts_comp, ts_diff, ts_filter;
	uint64_t ts_diff_abs, ts_filter_abs;
	
	uint16_t year_staf;
	uint8_t mon_staf, day_staf, h_staf, m_staf, s_staf, cs_staf;
	uint16_t prev_year_staf;
	uint8_t prev_mon_staf, prev_day_staf, prev_h_staf, prev_m_staf, prev_s_staf, prev_cs_staf;
	int year_comp, mon_comp, day_comp, h_comp, m_comp, s_comp, ms_comp;
	int prev_year_comp, prev_mon_comp, prev_day_comp, prev_h_comp, prev_m_comp, prev_s_comp, prev_ms_comp;
	int32_t h_diff;
	uint8_t m_diff, s_diff;
	uint16_t ms_diff;
	int32_t h_filter;
	uint8_t m_filter, s_filter;
	uint16_t ms_filter;
	
	uint8_t resync_state = STATE_INIT;
	
	
	if (argc<2)
	{
		fprintf(stderr,"%s port\n",(argc>0)?argv[0]:"staf-resync");
		return 1;
	}
	
	fd = stafConnect(argv[1]);
	if (fd<1)
	{
		fprintf(stderr,"%d: Can't connect to %s.\n",fd,argv[1]);
		return fd;
	}
	
	r=stafGetState(fd, NULL, &state);
	if (r)
	{
		fprintf(stderr,"%d: Can't get STAF state.\n",r);
		stafDisconnect(fd);
		return r;
	}
	
	r=stafGetCalibration(fd, NULL, &calibration,0);
	if (r)
	{
		fprintf(stderr,"%d: Can't get STAF calibration setting.\n",r);
		stafDisconnect(fd);
		return r;
	}
	printf("Current calibration setting: %"PRIi16"\n",calibration);
	
	while (1)
	{
		r2 = clock_gettime(CLOCK_REALTIME, &comptime);
		r = stafGetTime(fd, NULL, &cs_staf, &s_staf, &m_staf, &h_staf, &day_staf, &mon_staf, &year_staf);
		if (r)
		{
			fprintf(stderr,"%d: Can't get STAF time.\n",r);
			stafDisconnect(fd);
			return r;
		}
		if (r2)
		{
			fprintf(stderr,"%d: Can't get computer time.\n",r2);
			stafDisconnect(fd);
			return r2;
		}
		
		ts_comp = ((double)(comptime.tv_sec))+(0.000000001 * (double)(comptime.tv_nsec));
		comptm = localtime(&(comptime.tv_sec));
		year_comp = comptm->tm_year+1900;
		mon_comp  = comptm->tm_mon+1;
		day_comp  = comptm->tm_mday;
		h_comp    = comptm->tm_hour;
		m_comp    = comptm->tm_min;
		s_comp    = comptm->tm_sec;
		ms_comp   = (int)(comptime.tv_nsec/1000000);
		
		staftm.tm_year = (int)(year_staf-1900);
		staftm.tm_mon  = (int)(mon_staf-1);
		staftm.tm_mday = (int)day_staf;
		staftm.tm_hour = (int)h_staf;
		staftm.tm_min  = (int)m_staf;
		staftm.tm_sec  = (int)s_staf;
		staftm.tm_isdst = comptm->tm_isdst;
		staftime = timelocal(&staftm);
		ts_staf = ((double)staftime)+(0.01 * (double)cs_staf);
		
		ts_diff = ts_staf - ts_comp;
		ts_filter = (resync_state == STATE_INIT) ? ts_diff : (FILTER_FACTOR * ts_diff + (1.0 - FILTER_FACTOR) * ts_filter);
		
		ts_diff_abs = (uint64_t)(ts_diff * ((ts_diff>=0)?1000.0:-1000.0) + 0.5);
		ms_diff = (uint16_t)(ts_diff_abs % 1000);
		ts_diff_abs /= 1000;
		s_diff = (uint8_t)(ts_diff_abs % 60);
		ts_diff_abs /= 60;
		m_diff = (uint8_t)(ts_diff_abs % 60);
		ts_diff_abs /= 60;
		h_diff = (int32_t)((ts_diff >= 0) ? (ts_diff_abs) : (0-ts_diff_abs)); // will not give me minus zero...
		
		ts_filter_abs = (uint64_t)(ts_filter * ((ts_filter>=0)?1000.0:-1000.0) + 0.5);
		ms_filter = (uint16_t)(ts_filter_abs % 1000);
		ts_filter_abs /= 1000;
		s_filter = (uint8_t)(ts_filter_abs % 60);
		ts_filter_abs /= 60;
		m_filter = (uint8_t)(ts_filter_abs % 60);
		ts_filter_abs /= 60;
		h_filter = (int32_t)((ts_filter >= 0) ? (ts_filter_abs) : (0-ts_filter_abs));
		
		if (resync_state != STATE_INIT)
			printf("\e[5F");
		
		switch(resync_state)
		{
		case STATE_INIT:
			printf("Offset: %19.6f\n",ts_diff);
			printf("state: INIT                \n");
			if ((ts_diff > 2) || (ts_diff < -2))
				resync_state = STATE_WAIT_FULL_MINUTE;
			else
				resync_state = STATE_WAIT_FILTER;
			break;
			
		case STATE_WAIT_FULL_MINUTE:
			printf("state: WAIT FOR FULL MINUTE\n");
			if (s_comp < prev_s_comp)
				resync_state = STATE_SET_TIME;
			break;
			
		case STATE_SET_TIME:
			printf("state: SET TIME            \n");
			
			r = stafSetTime(fd, NULL, m_comp, h_comp, 0);
			if (r)
			{
				fprintf(stderr,"%d: Can't set STAF time.\n",r);
				stafDisconnect(fd);
				return r;
			}
			
			r = stafSetDate(fd, NULL, day_comp, mon_comp, 0);
			if (r)
			{
				fprintf(stderr,"%d: Can't set STAF time.\n",r);
				stafDisconnect(fd);
				return r;
			}
			
			r = stafSetYear(fd, NULL, year_comp, 0);
			if (r)
			{
				fprintf(stderr,"%d: Can't set STAF time.\n",r);
				stafDisconnect(fd);
				return r;
			}
			
			resync_state = STATE_WAIT_NEXT_MINUTE;
			cs_staf = 0;
			s_staf = 0;
			m_staf = m_comp;
			h_staf = h_comp;
			day_staf = day_comp;
			mon_staf = mon_comp;
			year_staf = year_comp;
			break;
			
		case STATE_WAIT_NEXT_MINUTE:
			printf("state: WAIT FOR NEXT MINUTE\n");
			if (((int)s_staf) - ((int)prev_s_staf) < -30)
				resync_state = STATE_WAIT_FILTER;
			break;
			
		case STATE_WAIT_FILTER:
			printf("state: WAIT FOR FILTER     \n");
			if (((ts_filter - ts_diff) < 0.01) && ((ts_filter - ts_diff) > -0.01) && (s_staf > 5))
				resync_state = STATE_SET_RECALIBRATION;
			break;
			
		case STATE_SET_RECALIBRATION:
			printf("state: SET RECALIBRATION   \n");
			
			// bit 2 at address 0x08 of the RTC enables calibration every second instead of every minute
			
			r = stafGetDebugValue(fd, NULL, 0x08, &debug_val, 0);
			if (r)
			{
				fprintf(stderr,"%d: Can't get debug value.\n",r);
				stafDisconnect(fd);
				return r;
			}
			
			debug_val |= 0x04;
			
			r = stafSetDebugValue(fd, NULL, 0x08, debug_val, 0);
			if (r)
			{
				fprintf(stderr,"%d: Can't set debug value.\n",r);
				stafDisconnect(fd);
				return r;
			}
			
			resync_state = STATE_RESYNCHRONISATION;
			break;
			
		case STATE_RESYNCHRONISATION:
			printf("state: RESYNCHRONISATION   \n");
			if ((ts_filter < STOP_CONDITION) && (ts_filter > (-1.0 * STOP_CONDITION)))
			{
				resync_state = STATE_STOP_RECALIBRATION;
				break;
			}
			
			r = stafSetCalibration(fd, NULL, calculate_factor(ts_filter), 0);
			if (r)
			{
				fprintf(stderr,"%d: Can't set STAF calibration setting.\n",r);
				stafDisconnect(fd);
				return r;
			}
			break;
			
		case STATE_STOP_RECALIBRATION:
			printf("state: STOP RECALIBRATION  \n");
			
			r = stafGetDebugValue(fd, NULL, 0x08, &debug_val, 0);
			if (r)
			{
				fprintf(stderr,"%d: Can't get debug value.\n",r);
				stafDisconnect(fd);
				return r;
			}
			
			debug_val &= ~0x04;
			
			r = stafSetDebugValue(fd, NULL, 0x08, debug_val, 0);
			if (r)
			{
				fprintf(stderr,"%d: Can't set debug value.\n",r);
				stafDisconnect(fd);
				return r;
			}
			
			r = stafSetCalibration(fd, NULL, calibration, 0);
			if (r)
			{
				fprintf(stderr,"%d: Can't set STAF calibration setting.\n",r);
				stafDisconnect(fd);
				return r;
			}
			
			resync_state = STATE_END;
			break;
			
		case STATE_END:
		default:
			printf("state: END                 \n");
			resync_state = STATE_EXIT;
		}
		
		printf(
			"STAF:  %04"PRIu16"-%02"PRIu8"-%02"PRIu8" %02"PRIu8":%02"PRIu8":%02"PRIu8".%02"PRIu8"0%16.3f\n",
			year_staf, mon_staf, day_staf, h_staf, m_staf, s_staf, cs_staf, ts_staf
		);
		printf(
			"COMP:  %04d-%02d-%02d %02d:%02d:%02d.%03d%16.3f\n",
			year_comp, mon_comp, day_comp, h_comp, m_comp, s_comp, ms_comp, ts_comp
		);
		printf(
			"DIFF:% 15"PRIi32":%02"PRIu8":%02"PRIu8".%03"PRIu16"%19.6f\n",
			h_diff, m_diff, s_diff, ms_diff, ts_diff
		);
		printf(
			"FILTER:% 13"PRIi32":%02"PRIu8":%02"PRIu8".%03"PRIu16"%19.6f\n",
			h_filter, m_filter, s_filter, ms_filter, ts_filter
		);
		
		if (resync_state == STATE_EXIT)
			break;
		
		prev_year_staf = year_staf;
		prev_mon_staf = mon_staf;
		prev_day_staf = day_staf;
		prev_h_staf = h_staf;
		prev_m_staf = m_staf;
		prev_s_staf = s_staf;
		prev_cs_staf = cs_staf;
		
		prev_year_comp = year_comp;
		prev_mon_comp = mon_comp;
		prev_day_comp = day_comp;
		prev_h_comp = h_comp;
		prev_m_comp = m_comp;
		prev_s_comp = s_comp;
		prev_ms_comp = ms_comp;
		
	}
	
	r=stafSetState(fd,NULL,&state);
	if (r)
	{
		fprintf(stderr,"%d: Can't restore STAF state.\n",r);
		stafDisconnect(fd);
		return r;
	}
	
	stafDisconnect(fd);
	return 0;
}

int16_t calculate_factor(double diff)
{
	double Y;
	Y = (diff >= 0.0) ? diff : -1.0 * diff;
	Y = sqrt(Y);
	Y *= 255.0;
	if (Y > 255.0)
		Y = 255.0;
	if (Y < 5.0)
		Y = 5.0;
	if (diff < 0.0)
		Y *= -1.0;
	return (int16_t)Y;
}
